from functools import partial


def iter_linked_list(attr, ll):
    while ll:
        yield ll
        ll = getattr(ll, attr, None)


iter_frame = partial(iter_linked_list, "f_back")
iter_traceback = partial(iter_linked_list, "tb_next")


def iter_stack(frame, tb):
    for t in reversed(list(iter_traceback(tb))):
        if t.tb_frame == frame:
            break
        yield t.tb_frame, t.tb_lineno

    for f in iter_frame(frame):
        yield f, f.f_lineno


def iter_cause(exc):
    while exc:
        if exc.__cause__:
            exc = exc.__cause__
            yield exc, True
        elif exc.__context__:
            exc = exc.__context__
            yield exc, False
        else:
            exc = None


def force_iterable(value, dict_as_not_iterable=False):
    if value is None:
        return ()
    if isinstance(value, dict):
        if dict_as_not_iterable:
            return (value,)
    if isinstance(value, str):
        return (value,)
    try:
        return iter(value)
    except TypeError:
        return (value,)



import asyncio
import logging
import signal
import sys
import threading

try:
    import nest_asyncio
except ImportError:
    nest_asyncio = None

from .utils import current_origin

log = logging.getLogger(__name__)


loops = {}


def get_loop():
    # Here we get the magic cookie of our current thread in current pid
    origin = current_origin()
    try:
        loop = asyncio.get_running_loop()
        if origin not in loops or loop.is_running():
            if nest_asyncio:
                nest_asyncio.apply(loop)
            else:
                raise ImportError(
                    "If you want to use kalong in an asyncio environment "
                    "or use recursive debugging please install nest_asyncio"
                )
    except RuntimeError:
        pass

    if origin not in loops:
        loops[origin] = asyncio.new_event_loop()

        if (
            sys.platform != "win32"
            and threading.current_thread() is threading.main_thread()
        ):
            loops[origin].add_signal_handler(signal.SIGTERM, stop)
    return loops[origin]


def close_loop():
    origin = current_origin()
    try:
        loops[origin].close()
    finally:
        loops.pop(origin, None)


def clean_loops():
    log.info(f"Cleaning at exit {len(loops)} event loops")
    if not loops:
        return
    for loop in loops.values():
        loop.close()


def stop():
    for task in asyncio.all_tasks():
        task.cancel()



import asyncio
import logging
import os
import webbrowser
import socket
from threading import Lock

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectorError

from .errors import NoServerFoundError
from .forking import forkserver
from .loops import get_loop
from .utils import current_origin, url

log = logging.getLogger(__name__)
fork_lock = Lock()

websockets = {}
sessions = {}

websocket_options = {"max_msg_size": 1024 * 1024 * 1024}


async def websocket():
    ws, _ = await websocket_state()
    return ws


async def websocket_state():
    # Here we get the magic cookie of our current thread in current pid
    origin = current_origin()
    if origin in websockets:
        log.debug(f"Found existing websocket {origin}")

        return websockets[origin], True

    sessions[origin] = sessions.get(origin, ClientSession())

    with fork_lock:
        try:
            ws = await sessions[origin].ws_connect(url("back"), **websocket_options)
            log.info("Found existing kalong server")
        except ClientConnectorError:
            # If there are no server available, fork one
            log.info("No kalong server, starting one")
            forkserver()
            for _ in range(500):
                await asyncio.sleep(0.05)
                try:
                    ws = await sessions[origin].ws_connect(
                        url("back"), **websocket_options
                    )
                    break
                except ClientConnectorError:
                    pass
            else:
                raise NoServerFoundError()

        if os.getenv("KALONG_URLSOCKET"):
            try:
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                sock.connect(os.getenv("KALONG_URLSOCKET"))
                sock.send(url("front").encode("utf-8"))
                sock.shutdown(socket.SHUT_WR)
            finally:
                sock.close()
        # webbrowser.open should be in the mutex too, it's not thread safe
        elif os.getenv("KALONG_NO_BROWSER") or not webbrowser.open(url("front")):
            log.warning(
                f"Please open your browser to the following url: {url('front')}"
            )

    websockets[origin] = ws
    return ws, False


async def close_websocket():
    origin = current_origin()
    log.debug(f"Closing websocket {origin}")
    ws = websockets.pop(origin, None)
    session = sessions.pop(origin, None)

    if ws:
        await ws.close()
    if session:
        await session.close()


def die():
    log.info("Dying, closing socket.")
    loop = get_loop()
    loop.run_until_complete(close_websocket())


async def adie():
    log.info("Dying asynchronously, closing socket.")
    await close_websocket()


async def close_all(closeables):
    if not closeables:
        return
    await asyncio.gather(
        *[closeable.close() for closeable in closeables.values()],
        return_exceptions=True,
    )


def clean_websockets():
    log.info(f"Cleaning at exit {len(websockets)} ws and {len(sessions)} sessions")
    asyncio.run(close_all(websockets))
    asyncio.run(close_all(sessions))



import sys

from .obj import safe_bool, walk_obj


class FakeSTD(object):
    def __init__(self, answer, type):
        self.answer = answer
        self.type = type

    def write(self, s):
        self.answer.append({"type": self.type, "text": s})

    def flush(self):
        pass


class capture_display(object):
    def __init__(self, answer):
        self.obj = None
        self.answer = answer

    def __enter__(self):
        sys.displayhook = self.hook
        return self

    def __exit__(self, exctype, excinst, exctb):
        sys.displayhook = sys.__displayhook__

    def hook(self, obj):
        if obj is not None:
            answer = walk_obj(obj, set())
            answer["truthiness"] = safe_bool(obj)
            self.answer.append(answer)
        self.obj = obj


class capture_std(object):
    def __init__(self, answer):
        self.answer = answer

    def __enter__(self):
        sys.stdout = FakeSTD(self.answer, "out")
        sys.stderr = FakeSTD(self.answer, "err")

    def __exit__(self, exctype, excinst, exctb):
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__


class capture_exception(object):
    def __init__(self, answer):
        self.answer = answer

    def __enter__(self):
        sys.excepthook = self.hook

    def __exit__(self, exctype, excinst, exctb):
        sys.excepthook = sys.__excepthook__

    def hook(self, type_, value, tb):
        from ..debugger import serialize_exception

        self.answer.append(serialize_exception(type_, value, tb))



import logging
import sys
import threading
from multiprocessing import process
from pathlib import Path

from .communication import communicate
from .stepping import steppings, stop_trace
from .utils.iterators import iter_frame
from .websockets import die

log = logging.getLogger(__name__)

kalong_dir = str(Path(__file__).resolve().parent)


def trace(origin, frame, event, arg):
    stepping = steppings.get(origin)
    if not stepping:
        return

    type = stepping.get("type")
    originalFrame = stepping.get("frame")
    baseFrame = stepping.get("base_frame", originalFrame)
    lno = stepping.get("lno")
    skip_frames = stepping.get("skip_frames")
    filename = frame.f_code.co_filename
    into_type = None
    pending = None

    if type.startswith("stepInto"):
        into_type = type.split("stepInto")[1].lower() or "skipcore"
        type = "stepInto"

    parent_frames = list(iter_frame(frame))
    below = parent_frames.index(baseFrame) if baseFrame in parent_frames else -1

    # When we are in _shutdown of thread, program is finished
    if (
        filename == threading.__file__
        and frame.f_code.co_name in ["_shutdown", "_bootstrap_inner"]
        or filename == process.__file__
        and frame.f_code.co_name == "_bootstrap"
    ):
        stop_trace(frame)
        die()
        return

    # If we are tracing, we continue tracing if this is not an exception
    if type == "trace" and event != "exception":
        return sys.gettrace()

    # If we are continuing, don't trace anything but current frame
    if type == "continue":
        if frame == originalFrame:
            if stepping.get("condition"):
                # Check the condition
                response = stepping["condition"]()
                answers = response["answer"]
                # Break if answer is truthy
                break_ = len(answers) > 0 and (
                    any(answer.get("truthiness", True) for answer in answers)
                )
                if not break_:
                    # Continue tracing current frame
                    return sys.gettrace()
                pending = response

            # Continue tracing current frame
            elif event != "exception":
                return sys.gettrace()
        else:
            # Stop tracing if below
            return

    # Don't trace importlib bootstrapping
    if filename == "<frozen importlib._bootstrap>":
        return

    # Don't trace self
    if filename.startswith(kalong_dir):
        return

    # Don't trace under frames if we are not stepping 'into'
    if event == "call":
        if type != "stepInto":
            return
        # or if we are stepping into public methods only
        if into_type == "public" and frame.f_code.co_name.startswith("_"):
            return
        # or if we are not all and stepping into core methods
        if into_type == "skipcore" and frame.f_code.co_name.startswith("__"):
            return

    if skip_frames and (
        (event in ["call", "line"] and below - 1 < skip_frames)
        or (event == "return" and below - 1 < skip_frames)
    ):
        return sys.gettrace()

    if (
        # Trace: This is an exception: stop
        type == "trace"
        # Continue: This is an exception at current frame: stop
        or type == "continue"
        # Step: Normal stepping
        or type == "step"
        # StepInto: Stepping normally in new frame as well
        or type == "stepInto"
        # StepUntil: Stepping only if line number is greater than previously
        or (type == "stepUntil" and (frame != originalFrame or frame.f_lineno > lno))
        # StepOut: Stepping only if we are returning or in the parent frame
        or (
            type == "stepOut"
            and (
                (event == "return" and frame == originalFrame)
                or event == "line"
                and frame == originalFrame.f_back
            )
        )
    ):
        if event == "return":
            frame.f_locals["__kalong_return_value__"] = arg
        elif event == "exception":
            frame.f_locals["__kalong_exception__"] = arg
        # Enter the websocket communication loop that pauses the execution
        communicate(frame, event, arg, pending)

    return sys.gettrace()



import logging
import os
from argparse import REMAINDER, ArgumentParser

defaults = {
    "server": False,
    "protocol": "http",
    "host": "localhost",
    "port": 59999,
    "front_host": "localhost",
    "front_port": 59999,
    "ws_host": "localhost",
    "ws_port": 59999,
    "log": "warn",
    "detached": False,
    "command": [],
    "inject": None,
    "urlsocket": False,
}


class Config:
    def __init__(self):
        self.__dict__.update(defaults)

    def get_parser(self):
        parser = ArgumentParser(description="Kalong cli")
        parser.add_argument(
            "--server",
            action="store_true",
            help="Launch the kalong server. This option is used by kalong itself",
        )
        parser.add_argument(
            "--protocol",
            default=self.protocol,
            help="Protocol for contacting kalong server",
        )
        parser.add_argument(
            "--host",
            default=self.host,
            help="Host of kalong server",
        )
        parser.add_argument(
            "--port",
            type=int,
            default=self.port,
            help="Port of kalong server",
        )
        parser.add_argument(
            "--front-host",
            default=self.front_host,
            help="Host of kalong frontend, defaults to host option",
        )
        parser.add_argument(
            "--front-port",
            type=int,
            default=self.front_port,
            help="Port of kalong frontend, defaults to port option",
        )
        parser.add_argument(
            "--ws-host",
            default=self.ws_host,
            help="Host of kalong websocket server, defaults to host option",
        )
        parser.add_argument(
            "--ws-port",
            type=int,
            default=self.ws_port,
            help="Port of kalong websocket server, defaults to port option",
        )
        parser.add_argument(
            "--log",
            choices=["critical", "error", "warning", "info", "debug"],
            default=self.log,
            help="Kalong verbosity",
        )
        parser.add_argument(
            "--detached",
            action="store_true",
            help="Server won't exit in this mode after last client disconnect",
        )
        parser.add_argument(
            "--inject",
            type=int,
            help="Pid of a running process in which a debugger will be "
            "injected with gdb. This needs a working gdb and ptrace enabled",
        )
        parser.add_argument(
            "--break-at-start",
            action="store_true",
            help="Break at the start of the python file",
        )
        parser.add_argument(
            "--urlsocket",
            type=str,
            help="Path of the socket into which to feed the url for docker browser opening",
        )
        parser.add_argument(
            "command",
            nargs=REMAINDER,
            help="A python file to trace with kalong and its arguments or "
            "nothing to launch a shell instead",
        )
        return parser

    def parse_args(self):
        args = self.get_parser().parse_args()
        if not args.front_host:
            args.front_host = args.host
        if not args.front_port:
            args.front_port = args.port
        if not args.ws_host:
            args.ws_host = args.host
        if not args.ws_port:
            args.ws_port = args.port
        return args

    def from_args(self):
        for name, arg in vars(self.parse_args()).items():
            setattr(self, name, arg)

    def from_env(self):
        for name, value in defaults.items():
            value = os.getenv(f"KALONG_{name.upper()}")
            if value:
                setattr(self, name, type(name)(value))

        if os.getenv("KALONG_HOST") and not os.getenv("KALONG_FRONT_HOST"):
            self.front_host = self.host
        if os.getenv("KALONG_PORT") and not os.getenv("KALONG_FRONT_PORT"):
            self.front_port = self.port
        if os.getenv("KALONG_HOST") and not os.getenv("KALONG_WS_HOST"):
            self.ws_host = self.host
        if os.getenv("KALONG_PORT") and not os.getenv("KALONG_WS_PORT"):
            self.ws_port = self.port

    def get_args_for_server(self):
        yield "--server"

        if self.detached:
            yield "--detached"

        for name, default in defaults.items():
            if name in ["server", "detached", "command"]:
                continue
            value = getattr(self, name)
            if value != default:
                pretty_name = name.replace("_", "-")
                yield f"--{pretty_name}={value}"

    @property
    def log_level(self):
        return getattr(logging, self.log.upper())

    def __str__(self):
        return str(self.__dict__)



"""A new take on debugging"""

__version__ = "0.7.0"
import os
import sys
from pathlib import Path
from subprocess import run

from .config import Config

config = Config()
config.from_env()


def breakpoint():
    from .stepping import add_step, start_trace, stop_trace

    frame = sys._getframe().f_back
    stop_trace(frame)
    add_step("step", frame)
    start_trace(frame)


def break_above(level):
    from .stepping import add_step, start_trace, stop_trace

    frame = sys._getframe()
    while level > 0 and frame.f_back:
        level += 1
        frame = frame.f_back

    stop_trace(frame)
    add_step("step", frame)
    start_trace(frame)


def start_trace(break_=False, full=False, skip_frames=None):
    from .stepping import add_step, start_trace

    frame = sys._getframe().f_back
    add_step(
        "stepInto" if break_ else ("trace" if full else "continue"),
        frame,
        skip_frames,
    )
    start_trace(frame)


def stop_trace():
    from .stepping import stop_trace

    frame = sys._getframe().f_back
    stop_trace(frame)


class trace:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        stop_trace()
        start_trace(**self.kwargs)

    def __exit__(self, *args):
        stop_trace()


def run_file(filename, *args, break_=True):
    # Cleaning __main__ namespace
    import __main__

    from .utils import fake_argv

    __main__.__dict__.clear()
    __main__.__dict__.update(
        {
            "__name__": "__main__",
            "__file__": filename,
            "__builtins__": __builtins__,
        }
    )
    with open(filename, "rb") as fp:
        statement = compile(
            fp.read(), os.path.abspath(os.path.normcase(filename)), "exec"
        )

    globals = __main__.__dict__
    locals = globals
    with fake_argv(filename, *args):
        with trace(break_=break_):
            exec(statement, globals, locals)


def shell():
    # Launch a shell
    from .communication import communicate

    frame = sys._getframe()
    # Enter the websocket communication loop that pauses the execution
    communicate(frame, "shell", [])


def main():
    os.environ["PYTHONBREAKPOINT"] = "kalong.breakpoint"
    config.from_args()

    if config.server:
        from .server import serve

        serve()

    elif config.inject:
        kalong_dir = Path(__file__).resolve().parent.parent
        gdb_command = ["gdb", "-p", str(config.inject), "-batch"] + [
            "-eval-command=call %s" % hook
            for hook in [
                "(int) PyGILState_Ensure()",  # Getting the GIL
                '(int) PyRun_SimpleString("'
                f"print('* Kalong injection from {os.getpid()} *');"
                "import sys;"  # Putting kalong project directory in sys path:
                f"sys.path.insert(0, '{kalong_dir}');"
                "import kalong;"  # Setting breakpoint:
                "kalong.break_above(2);"
                '")',
                # Releasing the GIL with the PyGILState_Ensure handle:
                "(void) PyGILState_Release($1)",
            ]
        ]
        print(f"Running: {' '.join(gdb_command)}")
        run(gdb_command)

    else:
        if config.command:
            run_file(*config.command, break_=config.break_at_start)
        else:
            shell()


if __name__ == "__main__":
    main()



import ctypes
from inspect import (
    getcomments,
    getmodule,
    getmro,
    getsourcefile,
    getsourcelines,
    iscode,
    signature,
)

from .doc_lookup import get_fqn, get_online_doc


class ObjCache(object):
    def __init__(self):
        self.cache = {}

    def register(self, obj):
        ident = id(obj)
        self.cache[ident] = obj
        return ident

    def get(self, ident):
        return self.cache[ident]

    def clear(self):
        self.cache = {}


obj_cache = ObjCache()


def walk_iterable(iterable, walked):
    return [walk_obj(obj, walked) for obj in iterable]


def walk_mapping(mapping, walked):
    return [
        {"key": walk_obj(key, walked), "value": walk_obj(val, walked)}
        for key, val in mapping.items()
    ]


def walk_obj(obj, walked):
    id = obj_cache.register(obj)
    if id in walked:
        # Don't walk circular dependencies
        return {"type": "obj", "value": safe_repr(obj), "id": id}
    walked.add(id)

    if any([isinstance(obj, list), isinstance(obj, set), isinstance(obj, tuple)]):
        return {
            "type": "iterable",
            "subtype": type(obj).__name__,
            "values": walk_iterable(obj, walked),
            "id": obj_cache.register(obj),
        }
    if isinstance(obj, dict):
        return {
            "type": "mapping",
            "subtype": type(obj).__name__,
            "values": walk_mapping(obj, walked),
            "id": obj_cache.register(obj),
        }
    return {"type": "obj", "value": safe_repr(obj), "id": id}


def get_bases(cls):
    return {
        "id": obj_cache.register(cls),
        "name": cls.__name__,
        "bases": [get_bases(base) for base in cls.__bases__],
    }


def has_mixins(bases):
    return len(bases["bases"]) > 1 or any(has_mixins(base) for base in bases["bases"])


def get_infos(obj):
    infos = {}
    infos["type"] = type(obj).__name__
    cls = obj if type(obj) is type else type(obj)
    try:
        infos["signature"] = obj.__name__ + str(signature(obj))
    except Exception:
        pass

    infos["fqn"] = get_fqn(obj)
    if infos["fqn"]:
        infos["online_doc"] = get_online_doc(infos["fqn"])
    elif type(obj) is not type:
        type_fqn = get_fqn(type(obj))
        if type_fqn:
            infos["online_doc"] = get_online_doc(type_fqn)

    try:
        infos["bases"] = get_bases(cls)
        if has_mixins(infos["bases"]):
            try:
                infos["mro"] = [
                    {"id": obj_cache.register(m), "type": m.__name__}
                    for m in (getmro(cls))
                ]
            except Exception:
                pass
    except Exception:
        pass

    try:
        infos["file"] = getsourcefile(obj)
    except Exception:
        infos["file"] = ""

    try:
        srcs, lno = getsourcelines(obj)
        infos["file"] += f":{lno}"
        infos["source_size"] = len(srcs)
    except Exception:
        pass

    try:
        infos["module"] += f":{getmodule(obj).__name__}"
    except Exception:
        pass

    infos["comments"] = getcomments(obj)

    infos["id"] = id(obj)

    return infos


def sync_locals(frame, f_locals):
    """
    This is a cpython hack to synchronize new locals back into
    the fast local slots.
    """
    for key, value in f_locals.items():
        frame.f_locals[key] = value
    try:
        ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))
    except Exception:
        pass


def get_code(obj):
    if iscode(obj):
        return obj
    if hasattr(obj, "__func__"):
        obj = obj.__func__
    for attr in ["__code__", "gi_code", "ag_code", "cr_code", "f_code"]:
        if hasattr(obj, attr):
            code = getattr(obj, attr)
            if hasattr(code, "co_code"):
                return code


def safe_getattr(obj, key, default, include_exc=False):
    try:
        return getattr(obj, key)
    except Exception as e:
        if include_exc:
            return f"{default} {e.__class__.__name__}: {e}"
        return default


def safe_repr(obj, default="<unrepresentable>", include_exc=False):
    try:
        return repr(obj)
    except Exception as e:
        if include_exc:
            return f"{default} {e.__class__.__name__}: {e}"
        return default


def safe_bool(obj):
    try:
        return bool(obj)
    except Exception:
        return False



import atexit
import linecache
import logging
import signal
import sys
import threading
from functools import partial
from pathlib import Path

from .loops import clean_loops
from .utils import USER_SIGNAL, current_origin
from .utils.iterators import iter_frame
from .websockets import clean_websockets

log = logging.getLogger(__name__)

steppings = {}
kalong_dir = str(Path(__file__).resolve().parent)


def add_step(type, frame, skip_frames=None, condition=None):
    origin = current_origin()
    current = steppings.get(origin)
    steppings[origin] = {
        "type": type,
        "frame": frame,
        "lno": frame.f_lineno,
    }
    if condition:
        steppings[origin]["condition"] = condition
    if skip_frames is not None:
        steppings[origin]["skip_frames"] = skip_frames
    if current:
        # Keep parent stepping in case of recursive debugging
        if "_parent" in current:
            steppings[origin]["_parent"] = current["_parent"]
        if "skip_frames" in current:
            steppings[origin]["skip_frames"] = current["skip_frames"]
        # Keep base frame for skip_frames
        if "base_frame" not in current:
            steppings[origin]["base_frame"] = current["frame"]
        else:
            steppings[origin]["base_frame"] = current["base_frame"]


def clear_step():
    origin = current_origin()
    steppings.pop(origin, None)


def start_trace(frame):
    from .tracing import trace

    origin = current_origin()
    log.debug(f"Starting trace for {origin}")

    linecache.checkcache()
    current_trace = partial(trace, origin)
    sys.settrace(current_trace)
    for f in iter_frame(frame):
        f.f_trace = current_trace


def stop_trace(frame):
    sys.settrace(None)
    for f in iter_frame(frame):
        f.f_trace = None
    log.debug("Stopping trace")


@atexit.register
def cleanup():
    log.debug("Cleaning up at exit")
    stop_trace(sys._getframe())
    clean_websockets()
    clean_loops()


if threading.current_thread() is threading.main_thread():
    user_signal_handler = None

    def handle_user_signal(*args, **kwargs):
        log.info("Handling breakpoint on user signal")
        frame = sys._getframe().f_back
        add_step("step", frame)
        start_trace(frame)

        if user_signal_handler not in (signal.SIG_IGN, signal.SIG_DFL):
            user_signal_handler(*args, **kwargs)

    user_signal_handler = signal.signal(USER_SIGNAL, handle_user_signal)



import os
import subprocess
import sys
from pathlib import Path

from . import config


def forkserver():
    """
    Fork a detached server.py server that will continue to run as long
    as it has clients.
    """
    # We force the PYTHONPATH here in case of injection from outside
    kalong_dir = Path(__file__).parent.parent
    env = dict(os.environ)
    env["PYTHONPATH"] = (
        f'{kalong_dir}:{env["PYTHONPATH"]}' if env.get("PYTHONPATH") else kalong_dir
    )

    popen_args = (
        {
            "creationflags": subprocess.DETACHED_PROCESS
            | subprocess.CREATE_NEW_PROCESS_GROUP
        }  # Test this
        if sys.platform[:3] == "win"
        else {"start_new_session": True}
    )
    server = subprocess.Popen(
        [sys.executable, "-m", "kalong", *config.get_args_for_server()],
        close_fds=True,
        env=env,
        **popen_args,
    )
    # Raise error here?
    return server.poll()



import asyncio
import json
import linecache
import logging
import sys
import threading

from aiohttp import WSMsgType

from . import config
from .debugger import (
    get_frame,
    get_title,
    serialize_answer,
    serialize_answer_recursive,
    serialize_diff_eval,
    serialize_exception,
    serialize_frames,
    serialize_inspect,
    serialize_inspect_eval,
    serialize_suggestion,
    serialize_table,
)
from .errors import SetFrameError
from .loops import get_loop
from .stepping import add_step, clear_step, stop_trace
from .utils import basicConfig, get_file_from_code
from .websockets import adie, die, websocket_state

log = logging.getLogger(__name__)
basicConfig(level=config.log_level)


def exception_handler(loop, context):
    logging.error(f"Loop exit: {context}")


def communicate(frame, event, arg, pending=None):
    loop = get_loop()

    loop.set_exception_handler(exception_handler)
    try:
        loop.run_until_complete(communication_loop(frame, event, arg, pending))
    except asyncio.CancelledError:
        log.info("Loop got cancelled")
        die()


async def init(ws, frame, event, arg):
    tb = arg[2] if event == "exception" else None

    await ws.send_json({"type": "SET_THEME", "theme": event})
    await ws.send_json({"type": "SET_TITLE", "title": get_title(frame, event, arg)})
    await ws.send_json(
        {
            "type": "SET_FRAMES",
            "frames": list(serialize_frames(frame, tb)) if event != "shell" else [],
        }
    )


async def handle_message(ws, data, frame, event, arg):
    tb = arg[2] if event == "exception" else None
    if data["type"] == "HELLO":
        await init(ws, frame, event, arg)
        response = {
            "type": "SET_INFO",
            "config": config.__dict__,
            "main": threading.current_thread() is threading.main_thread(),
        }

    elif data["type"] == "GET_FILE":
        filename = data["filename"]
        file = "".join(linecache.getlines(filename))
        if not file:
            file = get_file_from_code(frame, filename)
        response = {
            "type": "SET_FILE",
            "filename": filename,
            "source": file,
        }

    elif data["type"] == "SET_PROMPT" or data["type"] == "REFRESH_PROMPT":
        try:
            step_frame = get_frame(frame, data.get("frame"), tb)
            eval_fun = (
                serialize_inspect_eval
                if data.get("command") == "inspect"
                else serialize_diff_eval
                if data.get("command") == "diff"
                else serialize_table
                if data.get("command") == "table"
                else serialize_answer_recursive
                if data.get("command") == "recursive_debug"
                else serialize_answer
            )

            def get_response():
                return {
                    "type": "SET_ANSWER",
                    "key": data["key"],
                    "command": data.get("command"),
                    "frame": data.get("frame"),
                    **eval_fun(data["prompt"], step_frame),
                }

            if data.get("command") == "condition":
                add_step("continue", step_frame, condition=get_response)
                response = {
                    "type": "DO_COMMAND",
                    "command": "continue",
                    "frame": data.get("frame"),
                    "stop": True,
                }
            else:
                response = get_response()

        except SetFrameError as e:
            frame = e.frame
            event = e.event
            arg = e.arg

            await init(ws, frame, event, arg)

            response = {
                "type": "SET_ANSWER",
                "key": data["key"],
                "command": data.get("command"),
                "frame": data.get("frame"),
                "prompt": data["prompt"].strip(),
                "answer": "",
                "duration": 0,
            }

    elif data["type"] == "REQUEST_INSPECT":
        response = {
            "type": "SET_ANSWER",
            "key": data["key"],
            "command": data.get("command"),
            **serialize_inspect(data["id"], frame),
        }

    elif data["type"] == "REQUEST_SUGGESTION":
        response = {
            "type": "SET_SUGGESTION",
            **serialize_suggestion(
                data["prompt"],
                data["from"],
                data["to"],
                data["cursor"],
                get_frame(frame, data.get("frame"), tb),
            ),
        }

    elif data["type"] == "DO_COMMAND":
        command = data["command"]
        response = {"type": "ACK", "command": command}
        if command == "run":
            clear_step()
            stop_trace(frame)
        elif command == "stop":
            clear_step()
            stop_trace(frame)
            await adie()
        else:
            step_frame = get_frame(frame, data.get("frame"), tb)
            add_step(command, step_frame)
        response["stop"] = True

    else:
        raise ValueError(f"Unknown type {data['type']}")
    return response


async def communication_loop(frame_, event_, arg_, pending=None):
    frame = frame_
    event = event_
    arg = arg_

    ws, existing = await websocket_state()
    try:
        await ws.send_json({"type": "PAUSE"})
    except ConnectionResetError:
        log.info("Connection was reset")
        stop_trace(frame)
        return

    if existing:
        # If the socket is already opened we need to update client state
        await init(ws, frame, event, arg)
        if pending:
            # Start by sending pending message
            await ws.send_json(pending)
        # Otherwise if it's new, just wait for HELLO to answer current state

    stop = False
    try:
        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                data = json.loads(msg.data)
                try:
                    response = await handle_message(ws, data, frame, event, arg)
                except Exception as e:
                    log.error(f"Error handling message {data}", exc_info=e)
                    response = {
                        "type": "SET_ANSWER",
                        "prompt": data.get("prompt", "?").strip(),
                        "key": data.get("key"),
                        "command": data.get("command"),
                        "frame": data.get("frame"),
                        "answer": [serialize_exception(*sys.exc_info(), "internal")],
                    }
                log.debug(f"Got {data} answering with {response}")
                response["local"] = True

                if response.pop("recursive", False):
                    await ws.send_json({"type": "PAUSE", "recursive": True})
                    await init(ws, frame, event, arg)

                stop = response.pop("stop", False)

                try:
                    await ws.send_json(response)
                except ConnectionResetError:
                    break

                if stop:
                    break

            elif msg.type == WSMsgType.ERROR:
                log.error("WebSocket closed", exc_info=ws.exception())
                break
    except asyncio.CancelledError:
        log.error("WebSocket communication cancelled")

    # Browser exited, stopping debug if we are not stepping
    if not stop:
        stop_trace(frame)
        await adie()



import inspect
import json
import logging
from pathlib import Path

log = logging.getLogger(__name__)

try:
    with open(Path(__file__).parent / "lookup.json") as f:
        lookup = json.load(f)
except FileNotFoundError:
    log.warning("Can't access doc lookup file", exc_info=True)
    lookup = {}


def get_fqn(obj):
    mod = inspect.getmodule(obj)
    mod_name = getattr(mod, "__name__", "") if mod else ""
    obj_name = getattr(obj, "__qualname__", getattr(obj, "__name__", ""))
    if not mod_name or mod == obj or mod_name == "builtins":
        return obj_name
    return ".".join(s for s in (mod_name, obj_name) if s)


def get_online_doc(fqn):
    return lookup.get(fqn)



import linecache
import os
import re
import signal
import socket
import sys
import threading
import traceback
from pathlib import Path

from .. import config
from .iterators import iter_frame

try:
    from log_colorizer import basicConfig
except ImportError:
    from logging import basicConfig  # noqa

try:
    import uncompyle6
except ImportError:
    uncompyle6 = None
except Exception:
    print("Uncompyle crashed at import:")
    traceback.print_exc()
    uncompyle6 = None

ORIGIN_RE = re.compile(r"(.+)__(.+)--(.+)")


def current_origin():
    return f"{socket.getfqdn()}__{os.getpid()}--{threading.get_ident()}"


def parse_origin(origin):
    match = ORIGIN_RE.search(origin)
    if not match:
        raise ValueError(f"Invalid origin format {origin} must match {ORIGIN_RE}")
    fqdn, pid, ident = match.groups()

    return fqdn, int(pid), int(ident)


class fake_argv:
    def __init__(self, *args):
        self.args = args

    def __enter__(self):
        self._old_args = sys.argv
        sys.argv = self.args

    def __exit__(self, *err):
        sys.argv = self._old_args


USER_SIGNAL = signal.SIGILL if sys.platform == "win32" else signal.SIGUSR1


def url(side):
    origin = current_origin()
    host = config.front_host if side == "front" else config.host
    port = config.front_port if side == "front" else config.port
    protocol = config.protocol
    return f"{protocol}://{host}:{port}/{side}/{origin}"


def human_readable_side(side):
    return "browser" if side == "front" else "program"


def discompile(code):
    if not uncompyle6:
        return ""
    try:
        return uncompyle6.deparse_code2str(code)
    except Exception:
        return ""


def get_file_from_code(frame, filename):
    for frame in iter_frame(frame):
        co_filename = frame.f_code.co_filename or "<unspecified>"
        if co_filename == "<frozen importlib._bootstrap>":
            co_filename = os.path.join(
                os.path.dirname(linecache.__file__),
                "importlib",
                "_bootstrap.py",
            )
        fn = Path(co_filename)
        if not co_filename.startswith("<"):
            fn = fn.resolve()

        if filename == str(fn):
            return discompile(frame.f_code)


def cutter_mock(text, cursor):
    ch = cursor["ch"]
    new_text = []
    for lno, line in enumerate(text.split("\n")):
        new_line = []
        for c, char in enumerate(line):
            if char == "!":
                new_line.append("[0].")
                if lno == cursor["line"] and c <= ch:
                    cursor["ch"] += 3

            else:
                new_line.append(char)
        new_text.append("".join(new_line))

    return "\n".join(new_text), cursor


def universal_get(item, key):
    if isinstance(item, dict):
        return item.get(key, "___void___")

    if hasattr(item, "__getitem__") and key.isdigit():
        try:
            return item.__class__.__getitem__(item, int(key))
        except IndexError:
            return "___void___"
    try:
        return getattr(item, str(key), "___void___")
    except Exception as e:
        return e


def universal_travel(item, path):
    for key in path.split("."):
        item = universal_get(item, key)
        if item == "___void___":
            return "___void___"

    return item


def dedent(code):
    lines = code.split("\n")
    if not lines:
        return code
    indent = min(
        [len(line) - len(line.lstrip()) for line in lines if line.strip()],
        default=0,
    )
    return "\n".join(line[indent:] for line in lines)
